@@ -12,137 +12,15 @@ from glob import glob
 from tqdm import tqdm
 
 with boxx.inpkg():
+    from .feature_libs import CheckboardFeatureLib
     from .stereo import Stereo
     from .utils import r_t_to_T, intrinsic_format_conversion
     from . import utils
     from .__info__ import __version__
 
-
 TEMP = __import__("tempfile").gettempdir()
 
 
-class MetaFeatureLib:
-    def find_image_points(self, d):
-        """
-        d is a dict for each checkboard image, including keys like "img", "path"
-        
-        Please calculate d["image_points"] and d["object_points"] base on d["img"]
-    
-        The method should:
-            Set d["image_points"] as np.array of shape(n, 2) or {id: shape(n, 2)}
-            Set d["object_points"] as np.array of shape(n, 3) or {id: shape(n, 3)}
-            You could store some other important data in dict d
-        """
-        raise NotImplementedError()
-
-    def vis(self, d, cam=None):
-        """
-        d is a dict for each checkboard image, including keys like "img", "path"
-        
-        return:
-            vis: np.array(h, w, 3)
-        """
-        vis = d["img"].copy()
-        if cam is not None and "T" in d:
-            vis = utils.vis_T(d["T"], cam, vis)
-        image_points = d["image_points"]
-        if isinstance(image_points, dict):
-            image_points = np.concatenate(list(image_points.values()), 0)
-        vis = utils.vis_point_uvs(image_points, vis)
-        return vis
-
-
-class CheckboardFeatureLib(MetaFeatureLib):
-    def __init__(self, checkboard=(11, 8), size_mm=25):
-        self.checkboard = checkboard
-        self.size_mm = size_mm
-        self.object_points = np.zeros(
-            (self.checkboard[0] * self.checkboard[1], 3), np.float32
-        )
-        self.object_points[:, :2] = np.float32(
-            np.mgrid[: self.checkboard[0], : self.checkboard[1]].T.reshape(-1, 2)
-            * self.size_mm
-            / 1000
-        )
-
-    def find_image_points(self, d):
-        img = d["img"]
-        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
-        ret, corners = cv2.findChessboardCorners(gray, self.checkboard, None)
-        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.01)
-        if ret:
-            cv2.cornerSubPix(
-                gray, corners, (4, 4), (-1, -1), criteria,
-            )
-            d["corners"] = corners
-            d["image_points"] = corners[:, 0]
-            d["object_points"] = self.object_points
-
-    def vis(self, d, cam=None):
-        vis = d["img"].copy() if "img" in d else boxx.imread(d["path"])
-        if cam is not None and "T" in d:
-            vis = utils.vis_T(d["T"], cam, vis)
-        h, w = vis.shape[:2]
-        draw_subpix = h < 1500
-        if draw_subpix:
-            sub_size = 2
-            vis = cv2.resize(vis, (w * sub_size, h * sub_size))
-            cv2.drawChessboardCorners(
-                vis, self.checkboard, d["corners"] * sub_size, True
-            )
-        else:
-            cv2.drawChessboardCorners(vis, self.checkboard, d["corners"], True)
-        return vis
-
-
-class ArucoFeatureLib(MetaFeatureLib):
-    def __init__(self, occlusion=False, detector_parameters=None):
-        import cv2.aruco
-
-        self.occlusion = occlusion
-        self.detector_parameters = detector_parameters or {}
-        aruco_temp_str = "480 240 0 580 240 0 580 340 0 480 340 0 480 120 0 580 120 0 580 220 0 480 220 0 480 0 0 580 0 0 580 100 0 480 100 0 0 360 0 100 360 0 100 460 0 0 460 0 480 360 0 580 360 0 580 460 0 480 460 0 480 480 0 580 480 0 580 580 0 480 580 0 360 480 0 460 480 0 460 580 0 360 580 0 240 480 0 340 480 0 340 580 0 240 580 0 120 480 0 220 480 0 220 580 0 120 580 0 0 480 0 100 480 0 100 580 0 0 580 0 0 240 0 100 240 0 100 340 0 0 340 0 0 120 0 100 120 0 100 220 0 0 220 0 0 0 0 100 0 0 100 100 0 0 100 0 120 0 0 220 0 0 220 100 0 120 100 0 240 0 0 340 0 0 340 100 0 240 100 0 360 0 0 460 0 0 460 100 0 360 100 0 120 120 0 220 120 0 220 220 0 120 220 0 240 120 0 340 120 0 340 220 0 240 220 0 360 120 0 460 120 0 460 220 0 360 220 0 120 360 0 220 360 0 220 460 0 120 460 0 360 360 0 460 360 0 460 460 0 360 460 0 240 360 0 340 360 0 340 460 0 240 460 0 120 240 0 220 240 0 220 340 0 120 340 0 360 240 0 460 240 0 460 340 0 360 340 0"
-        self.all_object_points = np.float32(
-            np.array(boxx.findints(aruco_temp_str)).reshape(-1, 3) / 1000.0
-        )
-        self.object_points = dict(enumerate(self.all_object_points.reshape(-1, 4, 3)))
-        self.aruco_dict_idx = cv2.aruco.DICT_6X6_250
-
-    def find_image_points(self, d):
-        img = d["img"]
-        gray = img if img.ndim == 2 else cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
-        parameters = cv2.aruco.DetectorParameters_create()
-        self.detector_parameters.setdefault("polygonalApproxAccuracyRate", 0.008)
-        [
-            setattr(parameters, key, value)
-            for key, value in self.detector_parameters.items()
-        ]
-
-        d["corners"], d["ids"], rejectedImgPoints = cv2.aruco.detectMarkers(
-            gray, cv2.aruco.Dictionary_get(self.aruco_dict_idx), parameters=parameters
-        )
-        d["valid"] = d["ids"] is not None and (
-            len(d["ids"]) == len(self.object_points) or self.occlusion
-        )
-        if d["valid"]:
-            d["ids"] = d["ids"][:, 0] if d["ids"].ndim == 2 else d["ids"]
-            d["image_points"] = dict(
-                zip(d["ids"], [corner.squeeze() for corner in d["corners"]])
-            )
-            d["object_points"] = {id: self.object_points[id] for id in d["ids"]}
-
-    def vis(self, d, cam=None):
-        img = d["img"].copy() if "img" in d else boxx.imread(d["path"])
-        cv2.aruco.drawDetectedMarkers(img, d["corners"], d["ids"])
-        if cam is not None and len(d["corners"]):
-            rvec, tvec, _ = cv2.aruco.estimatePoseSingleMarkers(
-                d["corners"], 0.1, cam.K, cam.D
-            )
-            for i in range(rvec.shape[0]):
-                cv2.aruco.drawAxis(img, cam.K, cam.D, rvec[i], tvec[i], 0.05)
-        return img
-
-
 class Cam(dict):
     def __init__(
         self,
@@ -232,7 +110,12 @@ class Cam(dict):
                 print(self)
                 return True
         if do_cache:
-            pickle.dump(self, open(cache_path, "wb"))
+            try:
+                pickle.dump(self, open(cache_path, "wb"))
+            except TypeError:
+                feature_lib = self.__dict__.pop("feature_lib")
+                pickle.dump(self, open(cache_path, "wb"))
+                self.feature_lib = feature_lib
             print(self)
         return False
 
@@ -276,18 +159,21 @@ class Cam(dict):
         assert img.shape[:2][::-1] == self.xy
         d = dict(img=img)
         feature_lib.find_image_points(d)
-        if isinstance(d["image_points"], dict):
-            d["image_points"] = np.concatenate(
-                [d["image_points"][k] for k in sorted(d["image_points"])], 0
+        if "image_points" in d:
+            if isinstance(d["image_points"], dict):
+                d["image_points"] = np.concatenate(
+                    [d["image_points"][k] for k in sorted(d["image_points"])], 0
+                )
+                d["object_points"] = np.concatenate(
+                    [d["object_points"][k] for k in sorted(d["object_points"])], 0
+                )
+            retval, rvecs, tvecs, reprojectionError = cv2.solvePnPGeneric(
+                d["object_points"], d["image_points"][:, None], self.K, self.D
             )
-            d["object_points"] = np.concatenate(
-                [d["object_points"][k] for k in sorted(d["object_points"])], 0
+            T_board_in_cam = utils.r_t_to_T(rvecs[0], tvecs[0])
+            d.update(
+                T=T_board_in_cam, reprojectionError=reprojectionError, retval=retval
             )
-        retval, rvecs, tvecs, reprojectionError = cv2.solvePnPGeneric(
-            d["object_points"], d["image_points"][:, None], self.K, self.D
-        )
-        T_board_in_cam = utils.r_t_to_T(rvecs[0], tvecs[0])
-        d.update(T=T_board_in_cam, reprojectionError=reprojectionError, retval=retval)
         return d
 
     def project_points(self, xyzs, T=None):
@@ -327,6 +213,23 @@ class Cam(dict):
         depth_vis = np.uint8(cv2.applyColorMap(depth_uint8, cv2.COLORMAP_JET) * 0.75)
         return utils.vis_align(img, depth_vis)
 
+    def vis_image_points_cover(self):
+        # all_image_points = np.concatenate(self._get_points_for_cv2())
+        # vis = utils.vis_point_uvs(all_image_points, self.xy[::-1])
+        vis = self.xy[::-1]
+        for uvs in self._get_points_for_cv2():
+            vis = utils.vis_point_uvs(uvs, vis, convex_hull=True)
+
+        all_object_points = np.concatenate(self._get_points_for_cv2("object_points"))
+        object_points = np.unique(all_object_points, axis=0)
+        t_middle = np.mean(object_points, 0)
+        size_y = sorted(object_points.max(0) - object_points.min(0))[1]
+        for d in self.values():
+            if "T" in d:
+                T = d["T"] @ r_t_to_T(np.zeros((3,)), t_middle)
+                vis = utils.vis_T(T, self, vis, length=size_y / 2)
+        return vis
+
     def _get_points_for_cv2(self, name="image_points"):
         points = [self[key][name] for key in self.valid_keys]
         if isinstance(points[0], dict):
@@ -343,11 +246,19 @@ class Cam(dict):
 
     def __str__(self,):
         s = (
-            "Cam: \n\tname: '%s' \n\txy: %s \n\tfovs: %s \n\tvalid=%s/%s \n\tretval=%.2f \n\timage_path=%s \n"
+            "Cam: \n\tname: '%s' \n\txy: %s \n\tfovs: %s \n\tK: %s \n\tD: %s \n\tvalid=%s/%s \n\tretval=%.2f \n\timage_path=%s \n"
             % (
                 self.name,
                 self.xy,
                 utils._str_angle_dic(self.fovs),
+                "fx=%s, fy=%s, cx=%s, cy=%s"
+                % (
+                    round(self.fx, 1),
+                    round(self.fy, 1),
+                    round(self.cx, 1),
+                    round(self.cy, 1),
+                ),
+                str(self.D.round(2)),
                 len(self.__dict__.get("image_points", [])),
                 len(self),
                 self.__dict__.get("retval", -1),
@@ -407,6 +318,14 @@ class Cam(dict):
     def fy(self):
         return self.K[1, 1]
 
+    @property
+    def cx(self):
+        return self.K[0, 2]
+
+    @property
+    def cy(self):
+        return self.K[1, 2]
+
     @property
     def fovs(self):
         fovx = 2 * boxx.arctan(self.xy[0] / 2 / self.fx)
